﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using Generators;
using System.Threading;

namespace FractalGenerator
{
    public struct Options
    {
        public int xRes;
        public int yRes;
        public int AA;
        public int nThreads;
        public int historySize;

        public Options(int xres, int yres, int aa, int nthreads, int hsize)
        {
            xRes = xres;
            yRes = yres;
            AA = aa;
            nThreads = nthreads;
            historySize = hsize;
        }
    }

    public partial class FractalView : Form
    {
        private Options options;
        private IGenerator generator;

        private Rectangle zoomArea;
        private bool traceSel;
        private bool selecting;

        private List<Image> history;
        private int currentImage;

        public FractalView()
        {
            InitializeComponent();
        }

        public FractalView(string title, IGenerator generator, Options options)
        {
            this.options = options;
            this.generator = generator;

            this.traceSel = false;
            this.selecting = false;

            this.history = new List<Image>(options.historySize);
            this.currentImage = -1;

            InitializeComponent();

            Text = title;

            this.ClientSize = new Size(options.xRes, options.yRes + this.statusBar.Height + this.tsMenu.Height);
        }

        protected override void OnLoad(EventArgs e)
        {
            base.OnLoad(e);

            DrawFractal(this.generator.GetDomain());
        }

        private void pbxImage_MouseDown(object sender, MouseEventArgs e)
        {
            if (this.currentImage == this.history.Count - 1)
            {
                this.zoomArea = new Rectangle(e.X, e.Y, 0, 0);
                this.traceSel = true;
                this.selecting = true;
            }
        }

        private void pbxImage_MouseUp(object sender, MouseEventArgs e)
        {
            if (this.selecting)
            {
                if (e.X == this.zoomArea.X && e.Y == this.zoomArea.Y)
                    this.traceSel = false;
            }
            this.selecting = false;
        }

        private void pbxImage_Paint(object sender, PaintEventArgs e)
        {
            if (this.traceSel)
            {
                e.Graphics.DrawRectangle(new Pen(Brushes.White), this.zoomArea);
            }
            if (this.selecting)
            {
                Domain domain = ComputeDomain();
                Brush b = new SolidBrush(Color.White);
                
                string text = domain.ToString();
                Font f = new Font(FontFamily.GenericMonospace, 12.0f);

                SizeF s = e.Graphics.MeasureString(text, f);

                //Write domain
                e.Graphics.FillRectangle(Brushes.DarkGray, new Rectangle(10, 10, (int)s.Width + 6, (int)s.Height + 6));
                e.Graphics.DrawString(text, f, Brushes.White, new PointF(13, 13)); 
            }
        }

        private void pbxImage_MouseMove(object sender, MouseEventArgs e)
        {
            if (this.selecting)
            {
                int x1 = Math.Min(this.zoomArea.Left, e.X);
                int x2 = Math.Max(this.zoomArea.Left, e.X);
                int y1 = Math.Min(this.zoomArea.Top, e.Y);
                int y2 = Math.Max(this.zoomArea.Top, e.Y);

                this.zoomArea = new Rectangle(x1, y1, x2 - x1, y2 - y1);

                this.pbxImage.Invalidate();
            }
        }

        private void pbxImage_Click(object sender, EventArgs e)
        {
        }

        private void tsbZoom_Click(object sender, EventArgs e)
        {
            if (this.traceSel == false)
                return;

            DrawFractal(ComputeDomain());
        }

        private Domain ComputeDomain()
        {
            Domain domain = this.generator.GetDomain();

            Domain newDomain = new Domain();
            newDomain.X = domain.X + (domain.Width / this.options.xRes) * this.zoomArea.X;
            newDomain.Y = domain.Y - (domain.Height / this.options.yRes) * this.zoomArea.Y;
            newDomain.Width = domain.Width / (this.options.xRes / Math.Max(this.zoomArea.Width, 1));
            newDomain.Height = newDomain.Width * (domain.Height / domain.Width); //Keep aspect ratio

            return newDomain;
        }

        private void DrawFractal(Domain domain)
        {
            DateTime t1 = DateTime.Now;

            this.generator.SetDomain(domain);

            Cursor.Current = Cursors.WaitCursor;

            //Split work in tasks
            RenderTask[] tasks = new RenderTask[this.options.nThreads];
            Thread[] threads = new Thread[this.options.nThreads];

            double dy = domain.Height / this.options.nThreads;
            int dyres = this.options.yRes / this.options.nThreads;

            for (int i = 0; i < this.options.nThreads; ++i)
            {
                Domain dt = new Domain();
                dt.X = domain.X;
                dt.Y = domain.Y - (i * dy);
                dt.Width = domain.Width;
                dt.Height = dy;

                dyres = i < (this.options.nThreads - 1) ? dyres : this.options.yRes - dyres * i;

                tasks[i] = new RenderTask(this.generator, dt, this.options.xRes * this.options.AA, dyres * this.options.AA);
                threads[i] = new Thread(new ThreadStart(tasks[i].DoWork));
                threads[i].Start();
            }

            //Wait for the rendering tasks
            for(int i = 0; i < this.options.nThreads; ++i)
                threads[i].Join();

            //Compose the sections in one image
            Bitmap img = new Bitmap(this.options.xRes * this.options.AA, this.options.yRes * this.options.AA);
            Graphics gx = Graphics.FromImage(img);
            for (int i = 0; i < this.options.nThreads; ++i)
            {
                gx.DrawImage(tasks[i].Output, new Rectangle(0, dyres * i * this.options.AA, this.options.xRes * this.options.AA, dyres * this.options.AA));
                tasks[i].Output.Dispose();
            }

            if(this.options.AA > 1)
            {
                Bitmap res = new Bitmap(this.options.xRes, this.options.yRes);
                gx = Graphics.FromImage(res);
                gx.InterpolationMode = System.Drawing.Drawing2D.InterpolationMode.High;

                gx.DrawImage(img, new Rectangle(0, 0, this.options.xRes, this.options.yRes));

                img.Dispose();
                img = res;
            }

            //Apply gamma correction
            for(int x = 0; x < img.Width; ++x)
                for (int y = 0; y < img.Height; ++y)
                {
                    Color color = img.GetPixel(x, y);

                    double r = (float)color.R / 255;
                    double g = (float)color.G / 255;
                    double b = (float)color.B / 255;

                    if (r > 0.0031308) r = 1.055 * (Math.Pow(r, (1 / 2.2f))) - 0.055;
                    else r = 12.92 * r;
                    if (g > 0.0031308) g = 1.055 * (Math.Pow(g, (1 / 2.2f))) - 0.055;
                    else g = 12.92 * g;
                    if (b > 0.0031308) b = 1.055 * (Math.Pow(b, (1 / 2.2f))) - 0.055;
                    else b = 12.92 * b;

                    img.SetPixel(x, y, Color.FromArgb((int)(r * 255), (int)(g * 255), (int)(b * 255)));
                }

            DateTime t2 = DateTime.Now;

            this.tsslTime.Text = string.Format("Rendering: {0} s", t2.Subtract(t1).TotalSeconds);
            this.tsslDomain.Text = string.Format("Domain: {0}", this.generator.GetDomain().ToString());

            if (this.history.Count >= this.history.Capacity)
                this.history.RemoveAt(0);

            this.history.Add(img);

            this.currentImage = this.history.Count - 1;
            this.tsslIndex.Text = string.Format("Image {0}/{1}", this.currentImage + 1, this.history.Count);

            this.pbxImage.Image = img;

            this.traceSel = false;

            Cursor.Current = Cursors.Default;
        }

        private void tsbSave_Click(object sender, EventArgs e)
        {
            SaveFileDialog saveDlg = new SaveFileDialog();
            saveDlg.Filter = "jpeg format (*.jpg)|*.jpg";
            if (saveDlg.ShowDialog(this) == DialogResult.OK)
            {
                this.pbxImage.Image.Save(saveDlg.FileName, System.Drawing.Imaging.ImageFormat.Jpeg);
            }
        }

        private void tsbPrev_Click(object sender, EventArgs e)
        {
            if (this.currentImage > 0)
                this.currentImage--;

            this.pbxImage.Image = this.history[this.currentImage];
            this.tsslIndex.Text = string.Format("Image {0}/{1}", this.currentImage + 1, this.history.Count);
        }

        private void tsbNext_Click(object sender, EventArgs e)
        {
            if (this.currentImage < (this.history.Count - 1))
                this.currentImage++;

            this.pbxImage.Image = this.history[this.currentImage];
            this.tsslIndex.Text = string.Format("Image {0} /{1}", this.currentImage + 1, this.history.Count);
        }
    }

    class RenderTask
    {
        private IGenerator generator;
        public Domain Domain { get; set; }
        public int ResX {get; set;}
        public int ResY {get; set; }
        public Image Output { get; set; }

        protected RenderTask()
        {}
        public RenderTask(IGenerator generator, Domain d, int rx, int ry)
        {
            this.generator = generator;
            Domain = d;
            ResX = rx;
            ResY = ry;
        }

        public void DoWork()
        {
            Output = this.generator.Draw(Domain, ResX, ResY);
        }
    }
}
