﻿using System.Drawing;
using System;

using PolyLibTester;

namespace Generators
{
    public struct Domain
    {
        public double X { get; set; }
        public double Y { get; set; }
        public double Width { get; set; }
        public double Height { get; set; }

        public Domain(double x, double y, double w, double h) : this()
        {
            X = x;
            Y = y;
            Width = w;
            Height = h;
        }

        public override string ToString()
        {
            return string.Format("X={0:0.000}, Y={1:0.000}, W={2:0.000}, H={3:0.000}", X, Y, Width, Height);
        }
    }

    public interface IGenerator
    {
        IGenerator CreateGenerator();
        string GetName();
        Bitmap Draw(Domain domain, int resx, int resy);
        Domain GetDomain();
        void SetDomain(Domain domain);
    }

    public class Mandelbrot : IGenerator
    {
        #region Generator Members

        protected Domain domain;
        protected int maxIterations;

        public Mandelbrot()
        {
            this.domain = new Domain(-2.0f, 1.5f, 3.0f, 3.0f);
            this.maxIterations = 200;
        }

        public virtual IGenerator CreateGenerator()
        {
            return new Mandelbrot();
        }

        public virtual string GetName()
        {
            return "Mandelbrot";
        }

        public virtual Bitmap Draw(Domain domain, int resx, int resy)
        {
            Bitmap bmp = new Bitmap(resx, resy);

            double dx = domain.Width / resx;
            double dy = domain.Height / resy;

            for (int y = 0; y < resy; ++y)
            {
                double py = domain.Y - dy * y;
                for (int x = 0; x < resx; ++x)
                {
                    double px = domain.X + dx * x;

                    Color c = ComputePoint(new Complex(px, py));

                    bmp.SetPixel(x, y, c);
                }
            }
 
            return bmp;
        }

        public virtual Domain GetDomain()
        {
            return this.domain;
        }

        public virtual void SetDomain(Domain domain)
        {
            this.domain = domain;
        }

        #endregion

        protected virtual Color ComputePoint(Complex point)
        {
            int iterations = 0;

            Complex z = point;
            while ((Complex.ModulusSquared(z) < 4.0) && iterations < this.maxIterations )
            {
                iterations++;

                z = z * z + point;
            }

            if (iterations == this.maxIterations)
            {
                return Color.Black;
            }
            else
            {          
                return Shade(iterations, z);
            }
        }

        protected virtual Color Shade(int iterations, Complex z)
        {
            //mu = N + 1 - log (log  |Z(N)|) / log k
            double smooth = iterations + 1.0 - Math.Log(Math.Log(Complex.Modulus(z))) / Math.Log(2);

            double val = (smooth / maxIterations);

	        double dx=0.8;
            double ff = Clamp(val, 0.0, 1.0);   //clamp f in [0,1]
            double g = (float)(6.0 - 2.0 * dx) * val + dx;             //scale f to [dx, 6-dx]
            double R = Math.Max(0.0, (3.0 - Math.Abs(g - 4.0) - Math.Abs(g - 5.0)) / 2.0);
            double G = Math.Max(0.0, (4.0 - Math.Abs(g - 2.0) - Math.Abs(g - 4.0)) / 2.0);
            double B = Math.Max(0.0, (3.0 - Math.Abs(g - 1.0) - Math.Abs(g - 2.0)) / 2.0);

            return Color.FromArgb((int)(R * 255), (int)(G * 255), (int)(B * 255));
        }

        protected double Clamp(double val, double min, double max)
        {
            return val > min ? (val > max ? max : val) : min;
        }
    }

    public class MandelbrotEscapeTime : Mandelbrot
    {
        public MandelbrotEscapeTime()
            : base()
        {
        }

        public override IGenerator CreateGenerator()
        {
            return new MandelbrotEscapeTime();
        }

        public override string GetName()
        {
            return @"Mandelbrot ET";
        }

        protected override Color Shade(int iterations, Complex z)
        {
            double red = (iterations % 32) * 7;
            double green = (iterations % 16) * 14;
            double blue = (iterations % 128) * 2;

            return Color.FromArgb((int)red, (int)green, (int)blue);
        }
    }

    public class MandelbrotBW : Mandelbrot
    {
        public MandelbrotBW() : base()
        {
        }

        public override IGenerator CreateGenerator()
        {
            return new MandelbrotBW();
        }

        public override string GetName()
        {
            return @"Mandelbrot B/W";
        }

        protected override Color Shade(int iterations, Complex z)
        {
            int val = (int)(((double)iterations / (double)this.maxIterations) * 255);
            return Color.FromArgb(val, val, val);
        }
    }

    public class JuliaEscapeTime : IGenerator
    {
        #region Generator Members

        protected Domain domain;
        protected int maxIterations;

        public JuliaEscapeTime()
        {
            this.domain = new Domain(-1.5f, 1.5f, 3.0f, 3.0f);
            this.maxIterations = 200;
        }

        public virtual IGenerator CreateGenerator()
        {
            return new JuliaEscapeTime();
        }

        public virtual string GetName()
        {
            return "Julia ET";
        }

        public virtual Bitmap Draw(Domain domain, int resx, int resy)
        {
            Bitmap bmp = new Bitmap(resx, resy);

            double dx = domain.Width / resx;
            double dy = domain.Height / resy;

            for (int y = 0; y < resy; ++y)
            {
                double py = domain.Y - dy * y;
                for (int x = 0; x < resx; ++x)
                {
                    double px = domain.X + dx * x;

                    Color c = ComputePoint(new Complex(px, py));

                    bmp.SetPixel(x, y, c);
                }
            }

            return bmp;
        }

        public virtual Domain GetDomain()
        {
            return this.domain;
        }

        public virtual void SetDomain(Domain domain)
        {
            this.domain = domain;
        }

        #endregion

        protected virtual Color ComputePoint(Complex point)
        {
            int iterations = 0;

            Complex z = point;
            Complex k = new Complex(-0.74434, -0.10772);
            while ((Complex.ModulusSquared(z) < 4.0) && iterations < this.maxIterations)
            {
                iterations++;

                z = z * z + k;
            }

            if (iterations == this.maxIterations)
            {
                return Color.Black;
            }
            else
            {
                return Shade(iterations, z);
            }
        }

        protected virtual Color Shade(int iterations, Complex z)
        {
            double red = (iterations % 32) * 7;
            double green = (iterations % 16) * 14;
            double blue = (iterations % 128) * 2;

            return Color.FromArgb((int)red, (int)green, (int)blue);
        }

        protected double Clamp(double val, double min, double max)
        {
            return val > min ? (val > max ? max : val) : min;
        }
    }

    public class Newton3 : IGenerator
    {
        protected Domain domain;
        protected int maxIterations;
        protected double tol;
        protected Complex r1, r2, r3;

        public Newton3()
        {
            this.domain = new Domain(-2.0, 2.0, 4.0, 4.0);
            this.maxIterations = 50;
            this.tol = 0.0001;

            this.r1 = new Complex(1.0, 0.0);
            this.r2 = new Complex(-0.5, Math.Sin(2.0 * Math.PI / 3.0));
            this.r3 = new Complex(-0.5, -Math.Sin(2.0 * Math.PI / 3.0));
        }

        #region IGenerator Members

        public virtual IGenerator CreateGenerator()
        {
            return new Newton3();
        }

        public virtual string GetName()
        {
            return "Newton 3";
        }

        public virtual Bitmap Draw(Domain domain, int resx, int resy)
        {
            Bitmap bmp = new Bitmap(resx, resy);

            double dx = domain.Width / resx;
            double dy = domain.Height / resy;

            for (int y = 0; y < resy; ++y)
            {
                double py = domain.Y - dy * y;
                for (int x = 0; x < resx; ++x)
                {
                    double px = domain.X + dx * x;

                    Color c = ComputePoint(new Complex(px, py));

                    bmp.SetPixel(x, y, c);
                }
            }

            return bmp;
        }

        public virtual Domain GetDomain()
        {
            return this.domain;
        }

        public virtual void SetDomain(Domain domain)
        {
            this.domain = domain;
        }

        #endregion

        protected virtual Color ComputePoint(Complex point)
        {
            Complex z = point;
            int iterations = 0;
            while ((iterations < this.maxIterations) && (Complex.Abs(z - this.r1) >= this.tol) && (Complex.Abs(z - this.r2) >= this.tol) && (Complex.Abs(z - this.r3) >= this.tol))
            {
                iterations++;

                if (Complex.Abs(z) > 0)
                    z = z - (z * z * z - 1.0) / (z * z * 3.0);
            }

            return Shade(iterations, z, Complex.Abs(z - this.r1), Complex.Abs(z - this.r2), Complex.Abs(z - this.r3));
        }

        protected virtual Color Shade(int iterations, Complex z, double d1, double d2, double d3)
        {
            int val = (int)(((double)iterations / this.maxIterations) * 255);

            if (d1 < this.tol)
                return Color.FromArgb(val, 0, 0);
            if (d2 <= this.tol)
                return Color.FromArgb(0, val, 0);
            if (d3 <= this.tol)
                return Color.FromArgb(0, 0, val);

            return Color.Black;
        }
    }

    public class Newton3BW : Newton3
    {
        public override IGenerator CreateGenerator()
        {
            return new Newton3BW();
        }

        public override string GetName()
        {
            return @"Newton 3 B/W";
        }

        protected override Color Shade(int iterations, Complex z, double d1, double d2, double d3)
        {
            int val = (int)(((double)iterations / this.maxIterations) * 255);

            return Color.FromArgb(val, val, val);
        }
    }

    //Bifurcation Diagram Of Logistic Map
    public class BDOLM : IGenerator
    {
        #region Generator Members

        protected Domain domain;
        protected int maxIterations;

        public BDOLM()
        {
            this.domain = new Domain(2.5f, 1.0f, 1.5f, 1.0f);
            this.maxIterations = 200;
        }

        public virtual IGenerator CreateGenerator()
        {
            return new BDOLM();
        }

        public virtual string GetName()
        {
            return "Biforcation Diagram";
        }

        public virtual Bitmap Draw(Domain domain, int resx, int resy)
        {
            Bitmap bmp = new Bitmap(resx, resy);
            Graphics g = Graphics.FromImage(bmp);
            g.FillRectangle(Brushes.White, new Rectangle(0, 0, resx, resy));

            double dx = domain.Width / resx;
            double dy = domain.Height / resy;

            for (int i = 0; i < resx; ++i)
            {
                double r = domain.X + i * dx;
                double x = 0.5;

                for (int j = 0; j < 100; ++j)
                {
                    x = r * x * (1.0 - x);
                }
                for (int j = 0; j < 1000; ++j)
                {
                    x = r * x * (1.0 - x);
                    if (x < domain.Y && x > domain.Y - domain.Height)
                    {
                        double y = x - (domain.Y - domain.Height);
                        bmp.SetPixel(i, (resy - 1) - (int)(y * (resy / domain.Height)), Color.Black);
                    }
                }
            }

            return bmp;
        }

        public virtual Domain GetDomain()
        {
            return this.domain;
        }

        public virtual void SetDomain(Domain domain)
        {
            this.domain = domain;
        }

        #endregion
    }
}